/*-------------------------------------------------------------------*
 * scruples - screen compiler
 *
 * Converts a SPF-like panel into C code for use with fss
 *
 * Author: Tommy Sprinkle - December 2012
 *
 *-------------------------------------------------------------------*
 */

/*-----------------------------------------------------------------------------
|  Copyright (c) 2012-2013, Tommy Sprinkle (tommy@tommysprinkle.com)  
|
|  All rights reserved. Redistribution and use in source and binary forms, 
|  with or without modification, are permitted provided that the following 
|  conditions are met:
|
|    * Redistributions of source code must retain the above copyright notice, 
|      this list of conditions and the following disclaimer. 
|    * Redistributions in binary form must reproduce the above copyright  
|      notice, this list of conditions and the following disclaimer in the  
|      documentation and/or other materials provided with the distribution. 
|    * Neither the name of the author nor the names of its contributors may  
|      be used toendorse or promote products derived from this software  
|      without specific prior written permission. 
|
|   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS  
|   "AS IS" AND ANYEXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
|   LIMITED TO, THE IMPLIED WARRANTIESOF MERCHANTABILITY AND FITNESS  
|   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENTSHALL THE 
|   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
|   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
|   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
|   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
|   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
|   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  
|   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
|   OF THE POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


void Usage(char *programName)
{
	fprintf(stderr,"Usage:\n");
	fprintf(stderr,"%s input_file_name output_file_name\n",programName);
	fprintf(stderr,"\n  input_file_name   - panel to be compiled");
	fprintf(stderr,"\n  output_file_name  - file to contain generated C code\n");
}

/* returns the index of the first argument that is not an option; i.e.
   does not start with a dash or a slash
*/

int HandleOptions(int argc,char *argv[])
{
	int i,firstnonoption=0;

	for (i=1; i< argc;i++) {
		if (argv[i][0] == '/' || argv[i][0] == '-') {
			switch (argv[i][1]) {
				/* An argument -? means help is requested */
				case '?':
					Usage(argv[0]);
					break;
				case 'h':
				case 'H':
					if (!stricmp(argv[i]+1,"help")) {
						Usage(argv[0]);
						break;
					}
				default:
					fprintf(stderr,"unknown option %s\n",argv[i]);
					break;
			}
		}
		else {
			firstnonoption = i;
			break;
		}
	}
	return firstnonoption;
}

/*---------------------------------------------------------------------------*
 * char * trim(char *)
 *
 * trim trailing blanks from a string
 *---------------------------------------------------------------------------*
 */

char *trim(char *str)
{
	int i;

	i = strlen(str);				// Get length of string
	if(!i)							// Exit if length is zero
		return str;

	i--;							// Use as index into string

	while(i>=0 && str[i] == ' ')	// While trailing blank
	{
		str[i] = '\0';				// Terminate string here
		i--;
	}

	return str;						// Return pointed to trimmed string
}

/*---------------------------------------------------------------------------*
 * Define various phases (modes) of operation
 *---------------------------------------------------------------------------*
 */

#define modeINIT   1
#define modeATTR   2
#define modeBODY   3
#define modePROC   4
#define modeZVARS  5
#define modeZERROR 6

/*---------------------------------------------------------------------------*
 * Attribute Structure
 *
 * Defines characteristics of an Attribute Character
 *---------------------------------------------------------------------------*
 */

struct sAttrs
{
	char attr;						// The Attribute Character
	char type;						// Type of Attribute - "t" Text, "i" Input, "o" Output
	int  value;						// Actual bit value to be used for 3270 attribute fields
};

/*---------------------------------------------------------------------------*
 * Fields Structure
 *
 * Defines a field on the panel
 *---------------------------------------------------------------------------*
 */

struct sFields
{
	int   row;						// Row number 1-24
	int   col;						// Column     1-80
	int   attr;						// Actual bit value of 3270 Attribute fields
	int   type;						// Field Type - "t" Text, "i" Input, "o" Output
	int   len;						// Field Length
	char *name;						// Field Name
	char *text;						// Field Text Data (for Text fields)
};

/*---------------------------------------------------------------------------*
 * int isAttr(char attrChar, struct sAttrs *attrArray, int attrCount)
 *
 * determine if a character has been defined as an attribute character
 * Return 0 if not attribute character
 * Return Index Value +1 into array structure if attribute character
 *---------------------------------------------------------------------------*
 */

int isAttr(char c, struct sAttrs *pAttrs, int attrCnt)
{
	int i;

	for(i=0; i<attrCnt; i++)		// Loop through Attribute Structure Array
	{
		if(c == pAttrs[i].attr)		// Check to see if character matches
			return i+1;				// Return index+1 if a match
	}

	return 0;						// Return 0 if not found in array
}

#define ISATTR(c) isAttr(c, &attrs[0], attrCnt)

/*---------------------------------------------------------------------------*
 * Main Routine
 *
 * Everything Else Happens Here!!
 *---------------------------------------------------------------------------*
 */

int main(int argc,char *argv[])
{
	FILE *iFile;					// Input File Handle
	FILE *oFile;					// Output File Handle
	char *oFilePtr;                 // Output File Name Pointer

	time_t *timer;					// Execution Time Value

	int   i;						// General purpose variable
	int   mode;						// Current mode of operation
	int   InputActive;				// Still processing input
	int   a1;						// Attribute for current field being processed
	int   a2;						// Attribute of next field
	int   len;
	int   c1;						// Column offset of current field being processed

	int   Row;						// Current Row
	int   Col;						// Current Column

	char *p;						// Current offset into input buffer
	char *p1;						// Offset into input buffer for current field being processed
	char *text;						// Text contents from panel for current field being processed
	char *nullStr;					// A null String

	char  iBuf[256];				// Input Buffer

	struct sAttrs attrs[64];		// Attribute Array
	int    attrCnt;					// Number of active entries in Attribute Array

	struct sFields fields[1024];	// Field Array
	int    fldCnt;					// Number of active entries in Field Array

	if (argc == 1)					// If no arguments we call the Usage routine and exit
	{
		Usage(argv[0]);
		exit(8);
	}

	i = HandleOptions(argc,argv);	// Handle the program options

	if(i+1 > argc)					// Need two file names
	{
		Usage(argv[0]);
		exit(8);
	}


	iFile = fopen(argv[i],    "r");	// Open Input File
	if(!iFile)
	{
	   fprintf(stderr, "Input File [%s] Failed To Open\n",argv[i]);
       exit(8);
    }

	oFilePtr = argv[i+1];			// Save Pointer to Output File Name 

	nullStr = (char *) malloc(1);	// Create Null String
	*nullStr = '\0';

	fldCnt = 0;						// Set Field Count to Zero
	Row    = 0;
	Col    = 0;

	attrCnt = 3;					// Set Default Attribute Characters
	attrs[0].attr  = '%';
	attrs[0].value = 0x38;
	attrs[0].type  = 't';

	attrs[1].attr  = '+';
	attrs[1].value = 0x30;
	attrs[1].type  = 't';

	attrs[2].attr  = '_';
	attrs[2].value = 0x00;
	attrs[2].type  = 'i';


	mode        = modeINIT;			// Set initial mode
	InputActive = 1;				// Have not seen EOF

	while(InputActive)				// While more lines to read
	{
	   fgets(iBuf, 255, iFile);		// Read a line
	   for(i=0; i<256; i++)			// Trim off trailing \n
	   {
		   if(*(iBuf+i) == '\n')
		   {
			   *(iBuf+i) = '\0';
			   break;
		   }
	   }

	   //if(i >= 255)
	   //{
	   //   printf("Line Too Long... <<%s>>\n",iBuf);
	   //	exit(0);
	   //}

	   p = 0;						// Clear offset into Input Buffer

	   switch(mode)					// Branch depending on current Mode of Operation
	   {
	   case modeINIT:				// -------------------- Initilization Mode --------------------
          if(feof(iFile))			// Don't expect to find EOF here
		  {
		     fprintf(stderr, "Unexpected EOF\n");
		     exit(8);
		  }

		  if(!strncmp(")BODY", iBuf, 5))	// Begin )BODY Section
		  {
			  mode = modeBODY;				// Set new mode
			  break;
		  }

		  if(!strncmp(")ATTR", iBuf, 5))	// Begin )ATTR Section
		  {
			  mode = modeATTR;				// Set new mode
			  p = iBuf+5;					// point past ")ATTR"

			  while(*p == ' ')				// skip blanks
				  p++;

			  if(!*p)	        			//  Next if end-of-line
			     break;

			  if( strncmp("DEFAULT(", p, 8) )	// Override Default Characters?
			  {
				  fprintf(stderr, "Undefined Parameter on )ATTR Statement\n");
				  exit(8);
			  }

			  p += 8;						// skip over "DEFAULT("
			  i  = 0;						// index into Attribute Array

			  while(*p && *p != ')')		// looking for end of default definitions ")"
			  {
				  if(i > 2)
				  {
					  fprintf(stderr, "Too many Default Attribute Characters Specified <<%s>>\n",p);
					  exit(8);
				  }

				  attrs[i].attr = *p;		// Replace attribute character in array
				  i++;						// Next array entry
				  p++;						// Next character
			  } /* while(*p != ')') */
		  }
		  break;


      case modeATTR:						// -------------------- Attribute Character Definition Mode --------------------
         if(feof(iFile))					// Don't expect to find EOF here
         {
            printf("Unexpected EOF\n");
            exit(8);
         }

		 if(!strncmp(")BODY", iBuf, 5))		// Begin Body Section
		 {
			 mode = modeBODY;				// Set new mode
			 break;
		 }

		 p = iBuf;							// Start scanning input buffer

		 while(*p == ' ')					// Skip over blanks
			 p++;

		 if(!*p)							// Ignore blank lines
			 break;

		 attrs[attrCnt].attr  = *p;			// Save Attribute Character being defined
		 attrs[attrCnt].value = 0;			// Clear 3270 attribute bits
		 attrs[attrCnt].type  = ' ';        // Clear type
		 p++;

		 while(*p == ' ')					// skip blanks
			 p++;

		 if(!*p)							// Must have some definition
		 {
			 printf("Missing Attribute Values\n");
			 exit(8);
	     }

		 while(*p)							// While more data to process
		 {
	  	    if(!strncmp("TYPE(", p, 5))		// Handle TYPE(...)
		    {
				p += 5;						// skip over "TYPE("

				while(*p == ' ')			// skip blanks
					p++;

				if(!strncmp("INPUT", p, 5))			// Is it INPUT
			    {
					p += 5;							// skip over
					//attrs[attrCnt].value |= 0;	// no bits to set
					attrs[attrCnt].type = 'i';		// set type
			    }
				else if(!strncmp("OUTPUT", p, 6))	// Is it OUTPUT
				{
					p += 6;							// skip over
					attrs[attrCnt].value |= 0x30;	// set 3270 bits
					attrs[attrCnt].type = 'o';		// set type
			    }
				else if(!strncmp("TEXT", p, 4))		// Is it TEXT
				{
				   p += 4;							// skip over
		           attrs[attrCnt].value |= 0x30;	// set 3270 bits
				   attrs[attrCnt].type = 't';		// set type
		        }
				else								// Else - Invalid Type
				{
					fprintf(stderr, "Invalid value for TYPE <<%s>>\n", p);
					exit(8);
				}

		    }
		    else if(!strncmp("INTENS(", p, 7))		// Process Intensity
		    {
				p += 7;								// skip over

				while(*p == ' ')					// skip blanks
					p++;

				if(!strncmp("HIGH", p, 4))			// Is it HIGH
				{
					p += 4;							// skip over
					attrs[attrCnt].value |= 0x08;	// set bits
			    }
				else if(!strncmp("LOW", p, 3))		// Is it LOW
				{
					p += 3;							// skip over --- no 3270 bits to set
			    }
				else if(!strncmp("NON", p, 3))		// Is it NON-display
				{
					p += 3;							// skip over
					attrs[attrCnt].value |= 0x0C;	// set 3270 bits
				}
				else								// Else --- Invalid Value
				{
					fprintf(stderr, "Invalid Value for INTENS <<%s>>\n", p);
					exit(8);
			    }
		    }
		    else if(!strncmp("COLOR(", p , 6))		// Process COLOR
		    {
				p += 6;								// skip over
				while(*p == ' ')					// skip blanks
					p++;

				if(!strncmp("BLUE", p, 4))			// Blue?
				{
					p += 4;							// skip over
					i =  0xF1;						// Save 3270 value
				}
				else if(!strncmp("RED", p, 3))
				{
					p += 3;
					i =  0xF2;
				}
				else if(!strncmp("PINK", p ,4))
				{
					p += 4;
					i  = 0xF3;
				}
				else if(!strncmp("GREEN", p, 5))
				{
					p += 5;
					i  = 0xF4;
				}
				else if(!strncmp("TURQ", p, 4))
				{
					p += 4;
					i  = 0xF5;
				}
				else if(!strncmp("YELLOW", p ,6))
				{
					p += 6;
					i  = 0xF6;
				}
				else if(!strncmp("WHITE", p, 5))
				{
					p += 5;
					i  = 0xF7;
				}
				else								// Else --- Invalid Value for Color
				{
					fprintf(stderr, "Invalid Value for COLOR <<%s>>\n", p);
					exit(8);
				}

				attrs[attrCnt].value &= 0xFF00FF;	// Zero Color Attribute Bits
				attrs[attrCnt].value |= (i << 8);	// Set color attribute bits
		    }
		    else if(!strncmp("HILITE(", p, 7))		// Process Extended Highlighting
		    {
				p += 7;								// Skip over
				while(*p == ' ')					// skip blanks
					p++;

				if(!strncmp("BLINK", p, 5))			// Blink?
				{
					p += 5;							// skip over
					i  = 0xF1;						// save 3270 attribute bits
				}
			    else if(!strncmp("REVERSE", p, 7))	// Reverse ?
				{
					p += 7;							// Skip over
					i = 0xF2;						// save 3270 attribute bits
				}
				else if(!strncmp("USCORE", p, 6))	// Underscore ?
				{
					p += 6;							// Skip over
					i  = 0xF4;						// save 3270 attribute Bits
				}
				else								// Else --- Invalid Value
				{
					fprintf(stderr, "Invalid Value for HILITE <<%s>>\n", p);
					exit(8);
				}

				attrs[attrCnt].value &= 0xFFFF;		// Zero Extended Attribute Bits
				attrs[attrCnt].value |= (i << 16);	// Set extended attribute bits
		    }
		    else
		    {
				fprintf(stderr, "Invalid Parameter Specified for Attribute Definition <<%s>>\n",p);
				exit(8);
		    }

			while(*p && *p == ' ')					// skip blanks
				p++;

			if(*p != ')')							// Expect closing paren
			{
				fprintf(stderr, "Syntax Error on Attribute Definition\n");
				exit(8);
		    }

			p++;									// skip over ")"
			while(*p == ' ')						// skip blanks
				p++;
	     }

		 attrCnt++;									// increment attribute count
		 break;										// next input line



       case modeBODY:								// -------------------- BODY <> Define Fields Here --------------------
		 if(feof(iFile))							// Done if EOF
		 {
			 InputActive = 0;
			 break;
		 }

		 if(!strncmp(")PROC", iBuf, 5))				// Begin )PROC Section ?
		 {
			 mode = modePROC;						// Set New Mode
			 break;
		 }

		 if(!strncmp(")END", iBuf, 4))				// End
		 {
			 InputActive = 0;						// We can stop now
			 break;
		 }

		 Row++;										// Increment Row Number
		 p  = iBuf;									// First char of Input Buffer
		 p1 = 0;									// Not currently processing a field (offset)
		 c1 = 0;									// Not currently processing a field (attr char)

		 Col = 1; 									// Start in Column 1

		 do											// Scan up to 80 bytes
		 {
			 i = ISATTR(*p);						// Check to see if pointer to an Attribute Char
			 if( i || Col == 80 || *p == '\0' )		// If attr char or end of line...
			 {
				 a2 = i-1;							// Save as next attribute value
				 if(p1)								// Are we currently processing a field?
				 {
					 len = p - p1 - 1;				// Calc field length
					 if(len < 1 && attrs[a1].type != 't')  // Length of zero is invalid
					 {
						 fprintf(stderr, "Zero length field on Line %i at Column %i\n",Row,Col);
						 exit(8);
					 }

					 text = (char *) malloc(len+1);	// Save contents of field (either text or name)
					 memcpy(text, p1+1, len);
					 *(text+len) = '\0';
					 trim(text);					// trim off trailing blanks

					 switch(attrs[a1].type)			// switch based on current field attribute type
					 {
					 case 'i':						// Input or Output Field
					 case 'o':
						 if(strlen(text) > 8)		// Limit Field Name to 8 characters
						 {
							 fprintf(stderr, "Field Name too long: Row %i <<%s>>\n",Row,text);
							 exit(8);
						 }

						 fields[fldCnt].row  = Row;					// Save Row
						 fields[fldCnt].col  = c1;  				// Save Beginning Column
						 fields[fldCnt].attr = attrs[a1].value;		// Save 3270 attribute value
						 fields[fldCnt].type = attrs[a1].type;		// Save Field Type
						 fields[fldCnt].len  = len;					// Save Field Length
						 fields[fldCnt].name = text;				// Contents is Field Name
						 fields[fldCnt].text = nullStr;				// Null String for Text

						 fldCnt++;					// Increment Field Count
						 break;

					 case 't':						// Text Field
						 len = strlen(text);         				// Get trimmed length of text
						 if(len == 0)
							 break;									// Ignore Zero Length Text Strings
						 fields[fldCnt].row  = Row;					// Save Row
						 fields[fldCnt].col  = c1;				    // Save Beginning Column
						 fields[fldCnt].attr = attrs[a1].value;		// Save 3270 attribute value
						 fields[fldCnt].type = attrs[a1].type;		// Save Field Type
						 fields[fldCnt].len  = len;					// Save Field Length
						 fields[fldCnt].name = nullStr;				// Text fields do not have a name
						 fields[fldCnt].text = text;				// Field Text Contents
						 fldCnt++;					// Increment Field Count
						 break;

					 default:
					     fprintf(stderr, "Logic Error: Invalid Attribute Type\n");
						 exit(8);
				     } /* switch(attrs[a1].type) */


				 }  /* if(p1) */

				 c1 = Col+1;						// Start Column of Next Field (skip over attribute)
				 a1 = a2;							// Attribute for next field
				 p1 = p;							// Buffer offset for next field
			 } /* if( i = ISATTR(*p)) */

			 if(!*p || Col == 80)					// If End of Line
				 break;								// Break Out

			 p++;								    // next character
			 Col++;									// Increment Column Counter

	     } while(1);

		 break;




	   case modePROC:								// -------------------- PROC Section (.ZVARS) ----------
		 if(feof(iFile))							// Done if EOF
		 {
			 InputActive = 0;
			 break;
		 }


		 if(!strncmp(")END", iBuf, 4))				// Done if )END Statement
		 {
			 InputActive = 0;
			 break;
		 }

		 trim(iBuf);								// Trim Input Buffer

		 p = iBuf;
		 while(*p == ' ')							// skip blanks
			 p++;
		 if(!(*p))									// ignore blank lines
			 break;

		 if(strncmp(".ZVARS", p, 6))				// Look for .ZVARS
		 {
			 fprintf(stderr, "Invalid Statement in )PROC Section <<%s>>\n",p);
			 exit(8);
		 }

		 p += 6;									// skip over ".ZVARS"

		 while(*p == ' ')							// skip blanks
			 p++;

		 if(*p != '=')								// Expecting Equal Sign
		 {
			 fprintf(stderr, "Invalid Syntax on .ZVARS <<%s>>\n",p);
			 exit(8);
		 }

		 p++;										// skip over "="

		 while(*p == ' ')							// skip blanks
			 p++;

		 if(*p == '\'')								// ignore leading quote char
			 p++;

		 while(*p == ' ')							// skip blanks
			 p++;

		 if(*p != '(')								// expecting open paren "("
		 {
			 fprintf(stderr, "Invalid Syntax on .ZVARS <<%s>>\n",p);
			 exit(8);
		 }

		 p++;										// skip over "("


		 mode = modeZVARS;							// Set new Mode && Drop Into Next Section

		// *** NOTE *** no break statement here *** fall through

	   case modeZVARS:								// -------------------- ZVARS Processing Section --------------------
		 if(!p)										// If we didn't fall through...
			 p = iBuf;								// We  have to set buffer offset pointer


		 do											// DO until end of input line
		 {

		    while(*p == ' ')						// skip blanks
			    p++;

		    if(*p == ')')							// Check for closing paren ")"
		    {
		   	 mode = modeZERROR;
		   	 break;
		    }

		    p1 = p;									// save start position for Variable Name

		    while(*p && *p != ' ' && *p != ')')		// find end of variable name
			    p++;

		    len = p - p1;							// calc name length

		    if(len > 8)								// Max length of 8
		    {
		   	    fprintf(stderr, ".ZVAR Variable Name Too Long <<%s>>\n",p1);
		   	    exit(8);
		    }

		    text = (char *) malloc(len+1);			// allocate storage for name
		    memcpy(text, p1, len);					// copy in
		    *(text+len) = '\0';

			for(i=0;i<fldCnt;i++)					// find first "Z" field in array
			{
				if(!strcmp("Z", fields[i].name))
				{
					fields[i].name = text;			// replace field name
					break;
			    }
				if( i == fldCnt )
				{
					fprintf(stderr, ".ZVAR Variable Not Matched To Z Field <<%s>>\n",p1);
				    exit(8);
			    }
		    }

		 } while(*p);								// keep going until end of line

		 break;


	   case modeZERROR:								// -------------------- Should Not Find Anything After Here --------------------
		 if(!strncmp(")END", iBuf, 4))
		 {
		    InputActive = 0;
		    break;
		 }

	   	 printf("Invalid Statement in )Proc Section <<%s>>\n",iBuf);
		 exit(8);


	   } /* End Switch */

    } /* End While(InputActive) */

	oFile = fopen(oFilePtr,  "w");					// Open Output File
	if(!oFile)
	{
	   fprintf(stderr, "Output File [%s] Failed To Open\n",argv[i+1]);
       exit(8);
    }

	time(&timer);									// Get current time

	fprintf(oFile, "/* ---- Generated By SCRUPLES ----- */\n");
	strcpy(iBuf, "/* ----- ");
	strcat(iBuf, ctime(&timer) );
	iBuf[strlen(iBuf)-1] = '\0';
	strcat(iBuf, " ----- */\n\n");
	fprintf(oFile, iBuf);


	for(i=0;i<fldCnt;i++)							// Now we loop through the field array and generate output
	{
		switch(fields[i].type)						// switch based on field type
		{
		case 't':
			fprintf(oFile, "      fssTxt(%i, %i, 0X%6.6X, \"%s\");\n",
														 fields[i].row,
			                                             fields[i].col,
														 fields[i].attr,
														 fields[i].text);
			break;
		case 'i':
		case 'o':
			fprintf(oFile, "      fssFld(%i, %i, 0X%6.6X, \"%s\", %i, \"%s\");\n",
			                                             fields[i].row,
			                                             fields[i].col,
														 fields[i].attr,
														 fields[i].name,
														 fields[i].len,
														 fields[i].text);
			break;
		}
    }

	fprintf(oFile, "\n\n/* ----- End of Generated Code ----- */");

	fclose(iFile);									// Close Input File
	fclose(oFile);									// Close Output File

	printf("---------- SCRUPLES COMPLETE ----------\n");

	return 0;
}
